from __future__ import annotations

from typing import Any

from ikea_api.constants import Constants


class IkeaApi:
    def __init__(
        self,
        token: str | None = None,
        country_code: str = "ru",
        language_code: str = "ru",
    ):
        self._token_value = token

        Constants.COUNTRY_CODE, Constants.LANGUAGE_CODE = country_code, language_code

    def login(self, username: str, password: str):
        """
        Log in as registered user.
        Token expires in 24 hours.

        Since this method is using Selenium, you would have to wait up to 30 seconds.
        """

        from ikea_api.auth import get_authorized_token

        self._token = get_authorized_token(username, password)

    def login_as_guest(self):
        """Token expires in 30 days."""
        from ikea_api.auth import get_guest_token

        self._token = get_guest_token()

    def reveal_token(self):
        """Retrieve token."""
        return self._token_value

    @property
    def _token(self):
        if self._token_value is None:
            raise RuntimeError("No token provided")
        return self._token_value

    @_token.setter
    def _token(self, value: Any):
        self._token_value = value

    @property
    def Cart(self):
        """Manage cart."""
        if not hasattr(self, "_cart"):
            from ikea_api.endpoints import Cart

            self._cart = Cart(self._token)
        return self._cart

    def OrderCapture(self, zip_code: str, state_code: str | None = None):
        """Get available delivery services. Object is callable."""
        from ikea_api.endpoints import OrderCapture

        return OrderCapture(self._token, zip_code, state_code)()

    @property
    def Purchases(self):
        """Get information about your purchases."""
        if not hasattr(self, "_purchases"):
            from ikea_api.endpoints import Purchases

            self._purchases = Purchases(self._token)
        return self._purchases

    @property
    def fetch_items_specs(self):
        """Get information about items."""
        if not hasattr(self, "_fetch_items_specs"):
            from ikea_api.endpoints import fetch_items_specs

            self._fetch_items_specs = fetch_items_specs
        return self._fetch_items_specs



from __future__ import annotations

from enum import Enum
from json.decoder import JSONDecodeError
from typing import Any

from requests import Session

from ikea_api.constants import Constants
from ikea_api.errors import GraphqlError, IkeaApiError, UnauthorizedError


class Method(Enum):
    POST = "POST"
    GET = "GET"


class API:
    """Generic API class"""

    def __init__(self, token: str, endpoint: str):
        self._token, self._endpoint = token, endpoint

        self._session = Session()
        self._session.headers.update(
            {
                "Accept-Encoding": "gzip, deflate, br",
                "Accept-Language": Constants.LANGUAGE_CODE,
                "Connection": "keep-alive",
                "User-Agent": Constants.USER_AGENT,
                "Origin": Constants.BASE_URL,
                "Referer": Constants.BASE_URL + "/",
            }
        )
        if token is not None:
            self._session.headers["Authorization"] = "Bearer " + token

    def _error_handler(self, status_code: int, response: Any):
        pass

    def _basic_error_handler(self, status_code: int, response: Any | dict[str, Any]):
        if status_code == 401:  # Token did not passed
            raise UnauthorizedError(response)

        if "errors" in response:  # GraphQL error
            raise GraphqlError(response)

    def _call_api(
        self,
        endpoint: str | None = None,
        method: Method = Method.POST,
        headers: dict[str, str] | None = None,
        data: dict[Any, Any] | list[Any] | None = None,
    ):
        """Wrapper for request's post/get with error handling"""
        if not endpoint:
            endpoint = self._endpoint

        if method == Method.GET:
            response = self._session.get(endpoint, headers=headers, params=data)
        elif method == Method.POST:
            response = self._session.post(endpoint, headers=headers, json=data)

        try:
            response_dict: Any = response.json()
        except JSONDecodeError:
            raise IkeaApiError(response.status_code, response.text)

        self._basic_error_handler(response.status_code, response_dict)
        self._error_handler(response.status_code, response_dict)

        if not response.ok:
            raise IkeaApiError(response.status_code, response.text)

        return response_dict



from __future__ import annotations

import asyncio
import os
import time

from pyppeteer import launch
from pyppeteer.browser import Browser
from pyppeteer.page import Page

from ikea_api.api import API
from ikea_api.constants import Constants, Secrets


def get_guest_token():
    return GuestAuth().get_token()


class GuestAuth(API):
    def __init__(self):
        super().__init__(None, "https://api.ingka.ikea.com/guest/token")  # type: ignore
        self._session.headers.update(
            {
                "Accept": "*/*",
                "Accept-Language": "en-us",
                "X-Client-Id": Secrets.auth_guest_token_x_client_id,
                "X-Client-Secret": Secrets.auth_guest_token_x_client_secret,
            }
        )

    def get_token(self) -> str:
        response: dict[str, str] = self._call_api(
            data={"retailUnit": Constants.COUNTRY_CODE}
        )
        self._token = response["access_token"]
        return self._token


script = """
    chrome = { runtime: {} };
    const originalQuery = navigator.permissions.query;
    navigator.permissions.query = (parameters) =>
    parameters.name === "notifications"
        ? Promise.resolve({ state: Notification.permission })
        : originalQuery(parameters);
    navigator.plugins = [1, 2, 3, 4, 5];
    navigator.languages = ["en-US", "en"];
"""


async def _get_driver():
    return await launch(
        handleSIGINT=False,
        handleSIGTERM=False,
        handleSIGHUP=False,
        args=[
            "--no-sandbox",
            "--window-position=0,0",
            '--user-agent="Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_7) '
            "AppleWebKit/537.36 (KHTML, like Gecko) Chrome/92.0.4515.107 "
            'Safari/537.36"',
        ],
    )


# pyright: reportUnknownMemberType = false


async def _open_page(browser: Browser):
    page = await browser.newPage()
    await page.evaluateOnNewDocument(script)
    await page.goto(
        f"https://www.ikea.com/{Constants.COUNTRY_CODE}/{Constants.LANGUAGE_CODE}/profile/login/"
    )
    return page


async def _get_token_from_cookie(page: Page):
    token: str | None = None
    for _ in range(200):
        cookies = await page.cookies()
        for cookie in cookies:
            if cookie["name"] == "idp_reguser":
                token = cookie["value"]  # type: ignore
        if token:
            break
        time.sleep(0.1)
    return token


async def _fill_form(page: Page, username: str, password: str):
    usr_el, pswd_el = "input#username", "input#password"
    await page.click(usr_el)
    await page.type(usr_el, username)
    await page.click(pswd_el)
    await page.type(pswd_el, password)
    await page.click("button[type=submit]")


async def _save_error_screenshot(page: Page):
    fpath = os.path.abspath("login_error_screenshot.png")
    await page.screenshot(path=fpath)
    raise RuntimeError(f"Cannot log in. Take look at the screenshot: {fpath}")


async def _main(username: str, password: str):
    browser = await _get_driver()
    page = await _open_page(browser)
    await _fill_form(page, username, password)
    token = await _get_token_from_cookie(page)
    await browser.close()
    if not token:
        return await _save_error_screenshot(page)
    return token


def get_authorized_token(username: str, password: str):
    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    return loop.run_until_complete(_main(username, password))



from __future__ import annotations

from enum import Enum
from typing import Any

from ikea_api.api import API
from ikea_api.constants import Constants
from ikea_api.endpoints.purchases import queries as _queries


class OrderInfoQuery(Enum):
    StatusBannerOrder = "StatusBannerOrder"
    CostsOrder = "CostsOrder"
    ProductListOrder = "ProductListOrder"


class Purchases(API):
    def __init__(self, token: str):
        super().__init__(token, "https://purchase-history.ocp.ingka.ikea.com/graphql")
        origin = "https://order.ikea.com"
        self._session.headers.update(
            {
                "Accept": "*/*",
                "Accept-Language": Constants.LANGUAGE_CODE
                + "-"
                + Constants.COUNTRY_CODE,
                "Origin": origin,
                "Referer": f"{origin}/{Constants.COUNTRY_CODE}/{Constants.LANGUAGE_CODE}/purchases/",
            }
        )

    def _build_payload(
        self, operation_name: str | OrderInfoQuery, query: str, **variables: Any
    ):
        payload: dict[str, Any] = {
            "operationName": operation_name,
            "variables": {**variables},
            "query": query,
        }
        return payload

    def history(self, take: int = 5, skip: int = 0):
        """
        Get purchase history.

        Parameters are for pagination.
        If you want to see all your purchases set 'take' to 10000
        """
        payload = self._build_payload("History", _queries.history, take=take, skip=skip)
        return self._call_api(data=payload)

    def order_info(
        self,
        order_number: str | int,
        email: str | None = None,
        queries: list[OrderInfoQuery] = [
            OrderInfoQuery.StatusBannerOrder,
            OrderInfoQuery.CostsOrder,
            OrderInfoQuery.ProductListOrder,
        ],
        skip_products: int = 0,
        skip_product_prices: bool = False,
        take_products: int = 10,
    ):
        """
        Get order information: status and costs.

        :params order_number: ID of your purchase
        :params email: Your email. If set, there's no need to get token.
        :params queries: which queries to include in request
        :params skip_products: Relevant to ProductListOrder
        :params skip_product_prices: Relevant to ProductListOrder
        :params take_products: Relevant to ProductListOrder
        """
        order_number = str(order_number)
        headers = {"Referer": f"{self._session.headers['Origin']}/{order_number}/"}

        payload: list[dict[str, dict[str, Any]]] = []
        if OrderInfoQuery.StatusBannerOrder in queries:
            payload.append(
                self._build_payload(
                    "StatusBannerOrder",
                    _queries.status_banner_order,
                    orderNumber=order_number,
                )
            )
        if OrderInfoQuery.CostsOrder in queries:
            payload.append(
                self._build_payload(
                    "CostsOrder",
                    _queries.costs_order,
                    orderNumber=order_number,
                )
            )
        if OrderInfoQuery.ProductListOrder in queries:
            payload.append(
                self._build_payload(
                    "ProductListOrder",
                    _queries.product_list_order,
                    orderNumber=order_number,
                    receiptNumber="",
                    skip=skip_products,
                    skipPrice=skip_product_prices,
                    take=take_products,
                )
            )

        if email:
            headers["Referer"] = self._session.headers["Referer"] + "lookup"
            for d in payload:
                d["variables"]["liteId"] = email

        return self._call_api(headers=headers, data=payload)



from __future__ import annotations

from typing import Any


class IkeaApiError(Exception):
    """Generic API related exception."""


class GraphqlError(IkeaApiError):
    """Generic GraphQL exception"""

    def __init__(self, response: dict[str, Any]):
        self.errors: list[dict[str, Any]] = response["errors"]
        pretty_errors: list[str] = ["\\"]
        for e in self.errors:
            description = None
            if "path" in e:
                description = e["path"][0]
            elif "locations" in response:
                description = str(e["locations"][0])
            msg = f"{e['extensions']['code']} {e['message']}{': ' + description if description else ''}"
            pretty_errors.append(msg)
        super().__init__("\n".join(pretty_errors))


class UnauthorizedError(IkeaApiError):
    """Exception that is being called when cannot log in"""

    def __init__(self, response: dict[str, Any]):
        msg = response.get("moreInformation") or response.get("error") or response
        super().__init__(msg)


class ItemFetchError(IkeaApiError):
    ...


class OrderCaptureError(IkeaApiError):
    def __init__(self, response: dict[str, Any]):
        self.error_code = response["errorCode"]
        details = (
            ""
            if response["message"] == response["details"]
            else ": " + response["details"]
        )
        msg = f"{response['message']}{details}"
        super().__init__(msg)



from __future__ import annotations

import asyncio
import re
from asyncio.tasks import Task
from typing import Any

import aiohttp

from ikea_api.constants import Constants
from ikea_api.endpoints.item import build_headers, parse_item_code
from ikea_api.errors import ItemFetchError


def _async_fetch(urls: list[str | None], headers: dict[str, str]):
    async def main():
        async def fetch(session: aiohttp.ClientSession, url: str | None):
            if not url:
                return

            async with session.get(url=url, headers=headers) as response:
                try:
                    if response.status == 404:
                        raise ItemFetchError(response.url.__str__())
                    else:
                        return await response.json()
                except ItemFetchError as e:
                    url = e.args[0]
                    if not url:
                        return

                    opposite_url = build_opposite_url(url)
                    if not opposite_url:
                        return

                    r = await session.get(opposite_url, headers=headers)
                    if not r.ok:
                        raise ItemFetchError(parse_item_code(url))
                    return await r.json()

        async def fetch_all(session: aiohttp.ClientSession):
            tasks: list[Task[Any]] = []
            loop = asyncio.get_event_loop()
            for url in urls:
                task = loop.create_task(fetch(session, url))
                tasks.append(task)
            results = await asyncio.gather(*tasks)
            return results

        async with aiohttp.ClientSession() as session:
            res = await fetch_all(session)
            return res

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)
    res = loop.run_until_complete(main())
    return res


def build_url(item_code: str, is_combination: bool):
    if not len(item_code) == 8:
        return
    url = "{base_url}/{country}/{lang}/products/{folder}/{item_code}.json".format(
        base_url=Constants.BASE_URL,
        country=Constants.COUNTRY_CODE,
        lang=Constants.LANGUAGE_CODE,
        folder=item_code[5:],
        item_code="s" + item_code if is_combination else item_code,
    )
    return url


def build_opposite_url(url: str):
    match = re.findall(r"([sS])(\d{8})", url)
    if len(match) != 1:
        return
    match = match[0]
    len_ = len(match)
    is_combination = False
    if len_ == 1:
        item_code = match[0]
    elif len_ == 2:
        item_code = match[1]
        is_combination = True
    else:
        return
    return build_url(item_code, not is_combination)


def fetch(items: dict[str, bool]):
    # {'item_code': True (is SPR?)...}
    headers = build_headers({"Accept": "*/*"})
    headers.pop("Origin")
    urls: list[str | None] = [build_url(parse_item_code(i), items[i]) for i in items]
    responses = _async_fetch(urls, headers=headers)
    return [r for r in responses if r]



from __future__ import annotations

from ikea_api.endpoints.cart import Cart
from ikea_api.endpoints.item.item_ingka import fetch as _fetch_items_specs_ingka
from ikea_api.endpoints.item.item_iows import fetch as _fetch_items_specs_iows
from ikea_api.endpoints.item.item_pip import fetch as _fetch_items_specs_pip
from ikea_api.endpoints.order_capture import OrderCapture
from ikea_api.endpoints.purchases import Purchases


class fetch_items_specs:
    @staticmethod
    def ingka(items: str | list[str]):
        return _fetch_items_specs_ingka(items)

    @staticmethod
    def iows(items: str | list[str]):
        return _fetch_items_specs_iows(items)

    @staticmethod
    def pip(items: dict[str, bool]):
        return _fetch_items_specs_pip(items)


__all__ = ["Cart", "OrderCapture", "Purchases", "fetch_items_specs"]



from ikea_api.endpoints.cart.fragments import cart_props

cart = """
  query Cart(
    $languageCode: String
    ) {
    cart(languageCode: $languageCode) {
      ...CartProps
    }
  }
  %s
""" % (
    cart_props
)



from ikea_api.endpoints.cart.fragments import cart_props

add_items = """
  mutation AddItems(
    $items: [AddItemInput!]!
    $languageCode: String
  ) {
    addItems(items: $items, languageCode: $languageCode) {
      quantity
      context {
        userId
        isAnonymous
        retailId
      }
    }
  }
"""

clear_coupon = """
  mutation ClearCoupon(
    $languageCode: String
  ) {
    clearCoupon(languageCode: $languageCode) {
      ...CartProps
    }
  }

  %s
""" % (
    cart_props
)


clear_items = """
  mutation ClearItems(
    $languageCode: String
  ) {
    clearItems(languageCode: $languageCode) {
      ...CartProps
    }
  }

  %s
""" % (
    cart_props
)


copy_items = """
  mutation CopyItems(
    $sourceUserId: ID!
    $languageCode: String
  ) {
    copyItems(sourceUserId: $sourceUserId, languageCode: $languageCode) {
      ...CartProps
    }
  }

  %s
""" % (
    cart_props
)


remove_items = """
  mutation RemoveItems(
    $itemNos: [ID!]!
    $languageCode: String
  ) {
    removeItems(itemNos: $itemNos, languageCode: $languageCode) {
      ...CartProps
    }
  }

  %s
""" % (
    cart_props
)


set_coupon = """
  mutation SetCoupon(
    $code: String!
    $languageCode: String
  ) {
    setCoupon(code: $code, languageCode: $languageCode) {
      ...CartProps
    }
  }

  %s
""" % (
    cart_props
)


update_items = """
  mutation UpdateItems(
    $items: [UpdateItemInput!]!
    $languageCode: String
  ) {
    updateItems(items: $items, languageCode: $languageCode) {
      ...CartProps
    }
  }

  %s
""" % (
    cart_props
)



from __future__ import annotations

from functools import wraps
from typing import Any, Callable

from ikea_api.api import API
from ikea_api.constants import Constants, Secrets
from ikea_api.endpoints.cart import mutations, queries
from ikea_api.endpoints.item import parse_item_code


def _build_payload_and_call(func: Callable[..., Any]):
    @wraps(func)
    def inner(self: Cart, *args: Any, **kwargs: Any):
        res: str | tuple[str, dict[str, Any]] = func(self, *args, **kwargs)
        if isinstance(res, tuple):
            query, variables = res
        else:
            query, variables = res, {}

        return self._call_api(data=self._build_payload(query, **variables))  # type: ignore

    return inner


class Cart(API):
    def __init__(self, token: str):
        super().__init__(token, "https://cart.oneweb.ingka.com/graphql")
        self._session.headers["X-Client-Id"] = Secrets.cart_x_client_id

    def _build_payload(self, query: str, **variables: Any):
        payload = {
            "query": query,
            "variables": {"languageCode": Constants.LANGUAGE_CODE, **variables},
        }
        return payload

    def _make_templated_items(self, items: dict[str, int]):
        items_templated: list[dict[str, Any]] = []
        for item_code, qty in items.items():
            item_code = parse_item_code(item_code)
            if item_code:
                items_templated.append({"itemNo": item_code, "quantity": qty})
        return items_templated

    @_build_payload_and_call
    def show(self):
        return queries.cart

    @_build_payload_and_call
    def clear(self):
        return mutations.clear_items

    @_build_payload_and_call
    def add_items(self, items: dict[str, int]):
        """
        Add items to cart.
        Required items list format: {'item_no': quantity, ...}
        """
        items_templated = self._make_templated_items(items)
        return mutations.add_items, {"items": items_templated}

    @_build_payload_and_call
    def update_items(self, items: dict[str, int]):
        """
        Replace quantity for given item to the new one.
        Required items list format: {'item_no': quantity, ...}
        """
        items_templated = self._make_templated_items(items)
        return mutations.update_items, {"items": items_templated}

    @_build_payload_and_call
    def copy_items(self, source_user_id: str):
        """Copy cart from another account"""
        return mutations.copy_items, {"sourceUserId": source_user_id}

    @_build_payload_and_call
    def remove_items(self, item_codes: list[str]):
        """
        Remove items by item codes.
        """
        items_parsed = parse_item_code(item_codes)
        return mutations.remove_items, {"itemNos": items_parsed}

    @_build_payload_and_call
    def set_coupon(self, code: str):
        return mutations.set_coupon, {"code": code}

    @_build_payload_and_call
    def clear_coupon(self):
        return mutations.clear_coupon



#
# StatusBannerOrder
#

# also History
date_and_time = """
  fragment DateAndTime on DateAndTime {
    time
    date
    formattedLocal
    formattedShortDate
    formattedLongDate
    formattedShortDateTime
    formattedLongDateTime
  }
"""


delivery_date = """
  fragment DeliveryDate on DeliveryDate {
    actual {
      ...DateAndTime
    }
    estimatedFrom {
      ...DateAndTime
    }
    estimatedTo {
      ...DateAndTime
    }
    dateTimeRange
    timeZone
  }
"""


service_info = """
  fragment ServiceInfo on Service {
    id
    status
    date {
      ...DeliveryDate
    }
  }
"""


delivery_info = """
  fragment DeliveryInfo on DeliveryMethod {
    id
    serviceId
    status: status2
    type
    deliveryDate {
      ...DeliveryDate
    }
  }
"""


#
# CostsOrder
#

money = """
  fragment Money on Money {
    code
    value
  }
"""


tax_rate = """
  fragment TaxRate on TaxRate {
    percentage
    name
    amount {
      ...Money
    }
  }
"""

costs = """
  fragment Costs on Costs {{
    total {{
      ...Money
    }}
    delivery {{
      ...Money
    }}
    service {{
      ...Money
    }}
    discount {{
      ...Money
    }}
    sub {{
      ...Money
    }}
    tax {{
      ...Money
    }}
    taxRates {{
      ...TaxRate
    }}
  }}
  {}
  {}
""".format(
    money,
    tax_rate,
)


#
# ProductListOrder
#

article = """
  fragment Article on Product {
    name
    description
    href
    quantity
    decimalQuantity
    priceUnitText
    id
    splitDelivery
    image {
      small
      medium
      large
    }
  }
"""

product = """
  fragment Product on Product {
    ...Article
    unitPrice @skip(if: $skipPrice) {
      formatted
    }
    totalPrice @skip(if: $skipPrice) {
      formatted
    }
    assemblyRequired
  }

  %s
""" % (
    article
)



class Constants:
    USER_AGENT = (
        "Mozilla/5.0 (Macintosh; Intel Mac OS X 10_15_6) AppleWebKit/605.1.15"
        " (KHTML, like Gecko) Version/14.0.3 Safari/605.1.15"
    )
    BASE_URL = "https://www.ikea.com"
    LANGUAGE_CODE: str = "ru"
    COUNTRY_CODE: str = "ru"


class Secrets:
    cart_x_client_id = "66e4684a-dbcb-499c-8639-a72fa50ac0c3"
    item_ingka_x_client_id = "c4faceb6-0598-44a2-bae4-2c02f4019d06"
    item_iows_consumer = "MAMMUT#ShoppingCart"
    item_iows_contract = "37249"
    purchases_x_client_id = "af2525c3-1779-49be-8d7d-adf32cac1934"
    purchases_checkout_x_client_id = "6a38e438-0bbb-4d4f-bc55-eb314c2e8e23"
    auth_guest_token_x_client_id = "e026b58d-dd69-425f-a67f-1e9a5087b87b"  # nosec
    auth_guest_token_x_client_secret = (  # nosec
        "cP0vA4hJ4gD8kO3vX3fP2nE6xT7pT3oH0gC5gX6yB4cY7oR5mB"
    )



item_props = """
  fragment ItemProps on Item {
    itemNo
    quantity
    type
    fees {
      weee
      eco
    }
    isFamilyItem
    childItems {
      itemNo
    }
    regularPrice {
      unit {
        inclTax
        exclTax
        tax
        validFrom
        validTo
        isLowerPrice
        isTimeRestrictedOffer
        previousPrice {
          inclTax
          exclTax
          tax
        }
      }
      subTotalExclDiscount {
        inclTax
        exclTax
        tax
      }
      subTotalInclDiscount {
        inclTax
        exclTax
        tax
      }
      subTotalDiscount {
        amount
      }
      discounts {
        code
        amount
        description
        kind
      }
    }
    familyPrice {
      unit {
        inclTax
        exclTax
        tax
        validFrom
        validTo
      }
      subTotalExclDiscount {
        inclTax
        exclTax
        tax
      }
      subTotalInclDiscount {
        inclTax
        exclTax
        tax
      }
      subTotalDiscount {
        amount
      }
      discounts {
        code
        amount
        description
        kind
      }
    }
  }
"""


product_props = """
  fragment ProductProps on Product {
    name
    globalName
    isNew
    category
    description
    isBreathTaking
    formattedItemNo
    displayUnit {
      type
      imperial {
        unit
        value
      }
      metric {
        unit
        value
      }
    }
    unitCode
    measurements {
      metric
      imperial
    }
    technicalDetails {
      labelUrl
    }
    images {
      url
      quality
      width
    }
  }
"""

totals = """
  fragment Totals on Cart {
    regularTotalPrice {
      totalExclDiscount {
        inclTax
        exclTax
        tax
      }
      totalInclDiscount {
        inclTax
        exclTax
        tax
      }
      totalDiscount {
        amount
      }
      totalSavingsDetails {
        familyDiscounts
      }
    }
    familyTotalPrice {
      totalExclDiscount {
        inclTax
        exclTax
        tax
      }
      totalInclDiscount {
        inclTax
        exclTax
        tax
      }
      totalDiscount {
        amount
      }
      totalSavingsDetails {
        familyDiscounts
      }
      totalSavingsInclDiscount {
        amount
      }
    }
  }
"""


cart_props = f"""
  fragment CartProps on Cart {{
    currency
    checksum
    context {{
      userId
      isAnonymous
      retailId
    }}
    coupon {{
      code
      validFrom
      validTo
      description
    }}
    items {{
      ...ItemProps
      product {{
        ...ProductProps
      }}
    }}
    ...Totals
  }}
  {item_props}
  {product_props}
  {totals}
"""



"""IOWS Item API. Works only for Russian market"""

from __future__ import annotations

from typing import Any

from requests import Session

from ikea_api.constants import Constants, Secrets
from ikea_api.endpoints.item import generic_item_fetcher
from ikea_api.errors import ItemFetchError


def _build_url(items: dict[str, str]):
    endpoint = "https://iows.ikea.com/retail/iows/ru/ru/catalog/items/"
    templated_list: list[str] = []
    for item in items:
        templated_list.append(f"{items[item]},{item}")
    return endpoint + ";".join(templated_list)


def _fetch_items_specs(session: Session, input_items: list[str]):
    if len(input_items) == 0:
        return
    items: dict[str, str] = {}
    for item in input_items:
        items[item] = "ART"

    for i in range(3):
        url = _build_url(items)
        response = session.get(url)
        if i == 0 and len(items) == 1 and not response.ok:
            items[list(items.keys())[0]] = "SPR"
            url = _build_url(items)
            response = session.get(url)
            if not response.ok:  # TODO: Bad move
                raise ItemFetchError("Wrong Item Code")

        if not response.text:
            raise ItemFetchError("Wrong Item Code")
        r_json = response.json()

        if "RetailItemCommList" in r_json:
            if "RetailItemComm" in r_json["RetailItemCommList"]:
                return r_json["RetailItemCommList"]["RetailItemComm"]
            else:
                raise ItemFetchError(r_json)

        elif "RetailItemComm" in r_json:
            return [r_json["RetailItemComm"]]

        elif "ErrorList" in r_json:
            errors: list[dict[str, Any]] | dict[str, Any] = r_json["ErrorList"]["Error"]
            if not isinstance(errors, list):
                errors = [errors]

            for err in errors:
                for test in [
                    "ErrorCode" in err,
                    err["ErrorCode"]["$"] == 1101,
                    "ErrorMessage" in err,
                    "ErrorAttributeList" in err,
                    "ErrorAttribute" in err["ErrorAttributeList"],
                ]:
                    if not test:
                        raise ItemFetchError(err)

                attrs: dict[str, Any] = {}
                for attr in err["ErrorAttributeList"]["ErrorAttribute"]:
                    attrs[attr["Name"]["$"]] = attr["Value"]["$"]
                item_code = str(attrs["ITEM_NO"])
                item_type: str = attrs["ITEM_TYPE"]

                if i == 0:
                    items[item_code] = "SPR" if item_type == "ART" else "ART"
                elif i == 1:
                    items.pop(item_code)


def fetch(items: str | list[str]):
    headers = {
        "Accept": "application/vnd.ikea.iows+json;version=2.0",
        "Referer": f"{Constants.BASE_URL}/ru/ru/shoppinglist/",
        "Cache-Control": "no-cache, no-store",  # TODO: Does this required?
        "consumer": Secrets.item_iows_consumer,
        "contract": Secrets.item_iows_contract,
    }
    return generic_item_fetcher(items, headers, _fetch_items_specs, 90)



from ikea_api.endpoints.purchases.fragments import (
    costs,
    date_and_time,
    delivery_date,
    delivery_info,
    product,
    service_info,
)

history = """
  query History($skip: Int!, $take: Int!) {
    history(skip: $skip, take: $take) {
      id
      dateAndTime {
        ...DateAndTime
      }
      status
      storeName
      totalCost {
        code
        value
        formatted
      }
    }
  }

  %s
""" % (
    date_and_time
)


status_banner_order = f"""
  query StatusBannerOrder($orderNumber: String!, $liteId: String) {{
    order(orderNumber: $orderNumber, liteId: $liteId) {{
      id
      dateAndTime {{
        ...DateAndTime
      }}
      status
      services {{
        ...ServiceInfo
      }}
      deliveryMethods {{
        ...DeliveryInfo
      }}
    }}
  }}

  {date_and_time}
  {delivery_date}
  {service_info}
  {delivery_info}
"""


costs_order = """
  query CostsOrder($orderNumber: String!, $liteId: String) {{
    order(orderNumber: $orderNumber, liteId: $liteId) {{
      id
      costs {{
        ...Costs
      }}
    }}
  }}

  {}

""".format(
    costs,
)


product_list_order = """
  query ProductListOrder(
    $orderNumber: String!
    $liteId: String
    $skip: Int!
    $take: Int!
    $skipPrice: Boolean!
  ) {
    order(orderNumber: $orderNumber, liteId: $liteId) {
      id
      articles {
        ... on AnyDirectionProducts {
          quantity
          direction
          any(skip: $skip, take: $take) {
            ...Product
          }
        }
        ... on ExchangeProducts {
          numberOfInboundElements
          numberOfOutboundElements
          inboundQuantity
          outboundQuantity
          inbound(skip: $skip, take: $take) {
            ...Product
          }
          outbound(skip: $skip, take: $take) {
            ...Product
          }
        }
      }
    }
  }

  %s
""" % (
    product
)



from __future__ import annotations

from requests import Session

from ikea_api.constants import Constants, Secrets
from ikea_api.endpoints.item import generic_item_fetcher
from ikea_api.errors import ItemFetchError


def _fetch_items_specs(session: Session, items: list[str]):
    url = (
        "https://api.ingka.ikea.com/salesitem/communications/"
        + Constants.COUNTRY_CODE
        + "/"
        + Constants.LANGUAGE_CODE
    )
    params = {"itemNos": ",".join(items)}
    response = session.get(url, params=params)
    r_json = response.json()
    if "data" not in r_json and "error" in r_json:
        err_msg = None
        if "message" in r_json["error"]:
            error = r_json["error"]
            r_err_msg = error["message"]
            if r_err_msg == "no item numbers were found":
                try:
                    err_msg = error["details"][0]["value"]["keys"]
                except (KeyError, TypeError):
                    pass
            if not err_msg:
                err_msg = r_err_msg
        else:
            err_msg = r_json["error"]
        raise ItemFetchError(err_msg)

    return r_json


def fetch(items: str | list[str]):
    headers = {
        "Accept": "*/*",
        "Referer": f"{Constants.BASE_URL}/{Constants.COUNTRY_CODE}/{Constants.LANGUAGE_CODE}/order/delivery/",
        "x-client-id": Secrets.item_ingka_x_client_id,
    }
    return generic_item_fetcher(items, headers, _fetch_items_specs, 50)



from __future__ import annotations

import re
from typing import Any

from ikea_api.api import API, Method
from ikea_api.constants import Constants, Secrets
from ikea_api.endpoints.cart import Cart
from ikea_api.errors import IkeaApiError, OrderCaptureError


class OrderCapture(API):
    def __init__(self, token: str, zip_code: str | int, state_code: str | None = None):
        super().__init__(token, "https://ordercapture.ikea.ru/ordercaptureapi/ru")

        if Constants.COUNTRY_CODE != "ru":
            self._endpoint = (
                "https://ordercapture.ingka.com/ordercaptureapi/"
                + Constants.COUNTRY_CODE
            )

        zip_code = str(zip_code)
        validate_zip_code(zip_code)
        self._zip_code = zip_code

        validate_state_code(state_code)
        self._state_code = state_code

        self._session.headers["X-Client-Id"] = Secrets.purchases_x_client_id

    def __call__(self) -> list[dict[str, Any]]:
        return self.get_delivery_services()

    def _error_handler(self, status_code: int, response: dict[Any, Any]):
        if "errorCode" in response:
            raise OrderCaptureError(response)

    def _get_items_for_checkout_request(self):
        cart = Cart(self._token)
        cart_show = cart.show()
        items_templated: list[dict[str, Any]] = []
        try:
            if cart_show.get("data"):
                for d in cart_show["data"]["cart"]["items"]:
                    items_templated.append(
                        {
                            "quantity": d["quantity"],
                            "itemNo": d["itemNo"],
                            "uom": d["product"]["unitCode"],
                        }
                    )
        except KeyError:
            pass
        return items_templated

    def _get_checkout(self):
        """Generate checkout for items"""
        items = self._get_items_for_checkout_request()
        if len(items) == 0:
            return

        data = {
            "shoppingType": "ONLINE",
            "channel": "WEBAPP",
            "checkoutType": "STANDARD",
            "languageCode": Constants.LANGUAGE_CODE,
            "items": items,
            "deliveryArea": None,
        }

        response: dict[str, str] = self._call_api(
            endpoint=f"{self._endpoint}/checkouts",
            headers={"X-Client-Id": Secrets.purchases_checkout_x_client_id},
            data=data,
        )

        if "resourceId" in response:
            return response["resourceId"]
        raise IkeaApiError("No resourceId for checkout")

    def _get_delivery_area(self, checkout: str | None):
        """Generate delivery area for checkout from Zip Code and State Code"""
        data = {"enableRangeOfDays": False, "zipCode": self._zip_code}
        if self._state_code is not None:
            data["stateCode"] = self._state_code
        response = self._call_api(
            f"{self._endpoint}/checkouts/{checkout}/delivery-areas", data=data
        )

        if "resourceId" in response:
            return response["resourceId"]
        else:
            raise IkeaApiError("No resourceId for delivery area")

    def get_delivery_services(self):
        """Get available delivery services"""
        checkout: str | None = self._get_checkout()
        delivery_area = self._get_delivery_area(checkout)

        response = self._call_api(
            f"{self._endpoint}/checkouts/{checkout}/delivery-areas/{delivery_area}/delivery-services",
            method=Method.GET,
        )
        return response


def validate_zip_code(zip_code: str | int):
    if len(re.findall(r"[^0-9]", str(zip_code))) > 0:
        raise ValueError(f"Invalid zip code: {zip_code}")


def validate_state_code(state_code: str | None):
    if state_code is not None:
        if len(state_code) != 2 or len(re.findall(r"[^A-z]+", state_code)) > 0:
            raise ValueError(f"Invalid state code: {state_code}")



from __future__ import annotations

import re
from typing import Any, Callable, overload

from requests import Session

from ikea_api.constants import Constants

# TODO: Refactor all item endpoints


@overload
def parse_item_code(item_code: str | int) -> str:
    ...


@overload
def parse_item_code(item_code: list[str | int] | list[str]) -> list[str]:
    ...


def parse_item_code(item_code: str | int | list[str | int] | list[str]):
    def _parse(item_code: Any):
        found = re.search(r"\d{3}[, .-]{0,2}\d{3}[, .-]{0,2}\d{2}", str(item_code))
        res = ""
        if found:
            try:
                res = re.sub(r"[^0-9]+", "", found[0])
            except TypeError:
                pass
        return res

    err_msg = f"No items parsed: {str(item_code)}"
    if isinstance(item_code, list):
        item_code = list(set(item_code))
        res: list[str] = []
        for i in item_code:
            parsed = _parse(i)
            if parsed:
                res.append(parsed)
        if not res:
            raise ValueError(err_msg)
        return res
    else:
        parsed = _parse(item_code)
        if not parsed:
            raise ValueError(err_msg)
        return parsed


def build_headers(headers: dict[str, str]):
    new_headers = {
        "Origin": Constants.BASE_URL,
        "User-Agent": Constants.USER_AGENT,
        "Accept-Encoding": "gzip, deflate, br",
        "Accept-Language": Constants.LANGUAGE_CODE,
        "Connection": "keep-alive",
    }
    new_headers.update(headers)
    return new_headers


def generic_item_fetcher(
    items: str | list[str],
    headers: dict[str, str],
    func: Callable[..., Any],
    chunk_size: int,
):
    session = Session()
    session.headers.update(build_headers(headers))

    if isinstance(items, str):
        items = [items]

    items = [str(i) for i in items]
    items = list(set(items))
    items = parse_item_code(items)

    chunks = [items[x : x + chunk_size] for x in range(0, len(items), chunk_size)]
    responses: list[Any] = []
    for chunk in chunks:
        response = func(session, chunk)
        if isinstance(response, list):
            responses += response
        else:
            responses.append(response)
    return responses
